#include "sim-init.h"
#include "sim.h"
#include <linux/net.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <net/sock.h>

struct SimSocket
{};

static struct iovec *copy_iovec (const struct iovec *input, int len)
{
  int size = sizeof (struct iovec) * len;
  struct iovec *output = sim_malloc (size);
  sim_memcpy (output, input, size);
  return output;
}


int sim_sock_socket (int domain, int type, int protocol, struct SimSocket **socket)
{
  struct socket **kernel_socket = (struct socket **)socket;
  int retval = sock_create (domain, type, protocol, kernel_socket);
  return retval;
}
int sim_sock_close (struct SimSocket *socket)
{
  struct socket *kernel_socket = (struct socket *)socket;
  sock_release (kernel_socket);
  return 0;
}
static size_t iov_size(const struct msghdr *msg)
{
  size_t i;
  size_t size = 0;
  for (i = 0; i < msg->msg_iovlen; i++)
    {
      size += msg->msg_iov->iov_len;
    }
  return size;
}
ssize_t sim_sock_recvmsg (struct SimSocket *socket, struct msghdr *msg, int flags)
{
  struct socket *kernel_socket = (struct socket *)socket;
  struct iovec *kernel_iov = copy_iovec (msg->msg_iov, msg->msg_iovlen);
  struct iovec *user_iov = msg->msg_iov;
  msg->msg_iov = kernel_iov;
  int retval = sock_recvmsg (kernel_socket, msg, iov_size (msg), flags);
  msg->msg_iov = user_iov;
  sim_free (kernel_iov);
  return retval;
}
ssize_t sim_sock_sendmsg (struct SimSocket *socket, const struct msghdr *msg, int flags)
{
  struct socket *kernel_socket = (struct socket *)socket;
  struct iovec *kernel_iov = copy_iovec (msg->msg_iov, msg->msg_iovlen);
  struct msghdr kernel_msg = *msg;
  kernel_msg.msg_flags = flags;
  kernel_msg.msg_iov = kernel_iov;
  int retval = sock_sendmsg (kernel_socket, &kernel_msg, iov_size (msg));
  sim_free (kernel_iov);
  return retval;
}
int sim_sock_getsockname (struct SimSocket *socket, struct sockaddr *name, int *namelen)
{
  struct socket *sock = (struct socket *)socket;
  int retval = sock->ops->getname(sock, name, namelen, 0);
  return retval;
}
int sim_sock_getpeername (struct SimSocket *socket, struct sockaddr *name, int *namelen)
{
  struct socket *sock = (struct socket *)socket;
  int retval = sock->ops->getname(sock, name, namelen, 1);
  return retval;
}
int sim_sock_bind (struct SimSocket *socket, const struct sockaddr *name, int namelen)
{
  struct socket *sock = (struct socket *)socket;
  struct sockaddr_storage address;
  memcpy (&address, name, namelen);
  int retval = sock->ops->bind(sock, (struct sockaddr *)&address, namelen);
  return retval;
}
int sim_sock_connect (struct SimSocket *socket, const struct sockaddr *name, int namelen, int flags)
{
  struct socket *sock = (struct socket *)socket;
  struct sockaddr_storage address;
  memcpy (&address, name, namelen);
  int retval = sock->ops->connect(sock, (struct sockaddr *)&address, namelen, flags);
  return retval;
}
int sim_sock_listen (struct SimSocket *socket, int backlog)
{
  struct socket *sock = (struct socket *)socket;
  int retval = sock->ops->listen(sock, backlog);
  return retval;
}
int sim_sock_shutdown (struct SimSocket *socket, int how)
{
  struct socket *sock = (struct socket *)socket;
  int retval = sock->ops->shutdown(sock, how);
  return retval;
}
int sim_sock_accept (struct SimSocket *socket, struct SimSocket **new_socket, int flags)
{
  struct socket *sock, *newsock;
  int err;
  sock = (struct socket *)socket;

  // the fields do not matter here. If we could, 
  // we would call sock_alloc but it's not exported.
  err = sock_create_lite (0, 0, 0, &newsock);
  if (err < 0)
    {
      return err;
    }
  newsock->type = sock->type;
  newsock->ops = sock->ops;

  err = sock->ops->accept(sock, newsock, flags);
  if (err < 0)
    {
      sock_release (newsock);
      return err;
    }
  *new_socket = (struct SimSocket *)newsock;
  return 0;
}
int sim_sock_ioctl (struct SimSocket *socket, int request, char *argp)
{
  struct socket *sock = (struct socket *)socket;
  struct sock *sk;
  struct net *net;
  int err;

  sk = sock->sk;
  net = sock_net(sk);

  err = sock->ops->ioctl(sock, request, (long)argp);

  /*
   * If this ioctl is unknown try to hand it down
   * to the NIC driver.
   */
  if (err == -ENOIOCTLCMD)
    err = dev_ioctl(net, request, argp);
  return err;
}
int sim_sock_setsockopt (struct SimSocket *socket, int level, int optname,
			 const void *optval, int optlen)
{
  struct socket *sock = (struct socket *)socket;
  char *coptval = (char *)optval;
  int err;
  if (level == SOL_SOCKET)
    err = sock_setsockopt(sock, level, optname, coptval, optlen);
  else
    err = sock->ops->setsockopt(sock, level, optname, coptval, optlen);
  return err;
}
int sim_sock_getsockopt (struct SimSocket *socket, int level, int optname,
			 void *optval, int *optlen)
{
  struct socket *sock = (struct socket *)socket;
  int err;
  if (level == SOL_SOCKET)
    err = sock_getsockopt(sock, level, optname, optval, optlen);
  else
    err = sock->ops->getsockopt(sock, level, optname, optval, optlen);
  return err;
}
